module Hephaestus
  class Statistic
    def initialize(model)
      @model = model
      @associations = model.reflect_on_all_associations
    end

    def orphans(options = {})
      associations_of('BelongsToReflection').map do |association|
        next if options[:only_validated_associations] && has_validator?(association)
        { association: association.name, entries: @model.where("#{association.foreign_key} IS NULL") }
      end.compact
    end

    def empty_has_many(options = {})
      associations_of('HasManyReflection').map do |association|
        next if options[:only_validated_associations] && has_validator?(association)
        {
          association: association.name,
          entries: @model
                     .joins("LEFT OUTER JOIN #{association.table_name} ON #{model_table_name}.id = #{association.table_name}.#{association.foreign_key}")
                     .group("#{model_table_name}.id").having("count(#{association.table_name}.id) = 0")
        }
      end.compact
    end

    def empty_has_one(options = {})
      associations_of('HasOneReflection').map do |association|
        next if options[:only_validated_associations] && has_validator?(association)
        {
          association: association.name,
          entries: @model
                     .joins("LEFT OUTER JOIN #{association.table_name} ON #{association.table_name}.#{association.foreign_key} = #{model_table_name}.id")
                     .where("#{association.table_name}.id IS null")
        }
      end.compact
    end

    def empty_has_and_belongs_to_many(options = {})
      associations_of('HasAndBelongsToManyReflection').map do |association|
        next if options[:only_validated_associations] && has_validator?(association)
        {
          association: association.name,
          entries: @model
                     .joins("LEFT JOIN #{correct_join_table(association)} m ON #{model_table_name}.id = m.#{association.foreign_key}")
                     .where("m.#{association.foreign_key} IS NULL")
        }
      end.compact
    end

    def analyze_deviation(options = {})
      model_integer_columns.map do |column|
        values = @model.select("id, #{column}")
        find_deviation_anomalies(values, column, options) unless values.empty?
      end.first
    end

    def identify_clusters(relevant_columns, label_column, options = {})
      data = build_data(relevant_columns)
      labels = @model.pluck(label_column)
      best_result = nil

      (options[:maximum_clusters] || 8).times.with_index do |index|
        begin
          kmeans = KMeansClusterer.run(index+2, data, labels: labels, runs: 10)
          best_result = kmeans if best_result.nil? || (kmeans.silhouette > best_result.silhouette)
        rescue RuntimeError
          nil
        end
      end

      best_result
    end

    private

    def build_data(columns)
      @model.all.map do |record|
        columns.map do |column|
          if record.send(column).is_a?(TrueClass) || record.send(column).is_a?(FalseClass)
            bool_to_int(record.send(column))
          else
            record.send(column)
          end
        end
      end
    end

    def bool_to_int(bool)
      bool ? 1 : 0
    end

    def find_deviation_anomalies(values, column, options)
      normalized_values = values.map(&column.to_sym).compact

      unless normalized_values.empty?
        mean_plus_deviation = mean_plus_standard_deviation(normalized_values, options[:deviation_multiplier])
        anomalies = { column.to_sym => [] }
        values.each { |value| anomalies[column.to_sym] << { record: value, value: value.send(column) } if value.send(column).to_i > mean_plus_deviation }
        anomalies
      end
    end

    def mean_plus_standard_deviation(array, deviation_multiplier)
      mean = array.reduce(:+) / array.size
      deviation = 0
      array.each { |item| deviation += (item - mean)**2 }
      mean + (deviation_multiplier || 2)*Math.sqrt(deviation.to_f / array.size)
    end

    def model_integer_columns
      @model.columns.select { |column| (column.sql_type == 'int(11)' || column.sql_type.downcase == 'integer') && !column.name.include?('id') }.map(&:name)
    end

    def has_validator?(association)
      @model.validators.map(&:attributes).flatten.exclude?(association.foreign_key.to_sym) ||
        @model.validators.map(&:attributes).flatten.exclude?(association.foreign_key.delete('_id').to_sym) ||
          @model.validators.map(&:attributes).flatten.exclude?(association.foreign_key.pluralize.to_sym)
    end

    def correct_join_table(association)
      if ActiveRecord::Base.connection.table_exists?("#{association.table_name}_#{model_table_name}")
        "#{association.table_name}_#{model_table_name}"
      elsif ActiveRecord::Base.connection.table_exists?("#{model_table_name}_#{association.table_name}")
        "#{model_table_name}_#{association.table_name}"
      elsif ActiveRecord::Base.connection.table_exists?("#{model_table_name.singularize}_#{association.table_name}")
        "#{model_table_name.singularize}_#{association.table_name}"
      elsif ActiveRecord::Base.connection.table_exists?("#{model_table_name}_#{association.table_name.singularize}")
        "#{model_table_name}_#{association.table_name.singularize}"
      end
    end

    def model_table_name
      @model.table_name
    end

    def associations_of(type)
      @associations.select { |association| association.class.to_s.include?(type) }
    end
  end
end
